/*
	pr->pr_cmds.c

	@description@

	Copyright (C) 1996-1997  Id Software, Inc.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	See the GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to:

		Free Software Foundation, Inc.
		59 Temple Place - Suite 330
		Boston, MA  02111-1307, USA

	$Id$
*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <string.h>

#include "QF/console.h"
#include "QF/compat.h"
#include "QF/sys.h"
#include "QF/cmd.h"
#include "QF/va.h"
#include "host.h"
#include "world.h"
#include "QF/msg.h"
#include "server.h"
#include "sv_progs.h"

#define	RETURN_EDICT(p,e) (p->pr_globals[OFS_RETURN].int_var = EDICT_TO_PROG(p, e))

/*
===============================================================================

						BUILT-IN FUNCTIONS

===============================================================================
*/

char       *
PF_VarString (progs_t * pr, int first)
{
	int         i;
	static char out[256];

	out[0] = 0;
	for (i = first; i < pr->pr_argc; i++) {
		strcat (out, G_STRING (pr, (OFS_PARM0 + i * 3)));
	}
	return out;
}


/*
=================
PF_errror

This is a TERMINAL error, which will kill off the entire server.
Dumps self.

error(value)
=================
*/
void
PF_error (progs_t * pr)
{
	char       *s;
	edict_t    *ed;

	s = PF_VarString (pr, 0);
	Con_Printf ("======SERVER ERROR in %s:\n%s\n",
				PR_GetString (&sv_pr_state, pr->pr_xfunction->s_name), s);
	ed = PROG_TO_EDICT (pr, *sv_globals.self);
	ED_Print (pr, ed);

	Host_Error ("Program error");
}

/*
=================
PF_objerror

Dumps out self, then an error message.  The program is aborted and self is
removed, but the level can continue.

objerror(value)
=================
*/
void
PF_objerror (progs_t * pr)
{
	char       *s;
	edict_t    *ed;

	s = PF_VarString (pr, 0);
	Con_Printf ("======OBJECT ERROR in %s:\n%s\n",
				PR_GetString (&sv_pr_state, pr->pr_xfunction->s_name), s);
	ed = PROG_TO_EDICT (pr, *sv_globals.self);
	ED_Print (pr, ed);
	ED_Free (pr, ed);

	Host_Error ("Program error");
}



/*
==============
PF_makevectors

Writes new values for v_forward, v_up, and v_right based on angles
makevectors(vector)
==============
*/
void
PF_makevectors (progs_t * pr)
{
	AngleVectors (G_VECTOR (pr, OFS_PARM0), *sv_globals.v_forward,
				  *sv_globals.v_right, *sv_globals.v_up);
}

/*
=================
PF_setorigin

This is the only valid way to move an object without using the physics of the world (setting velocity and waiting).  Directly changing origin will not set internal links correctly, so clipping would be messed up.  This should be called when an object is spawned, and then only if it is teleported.

setorigin (entity, origin)
=================
*/
void
PF_setorigin (progs_t * pr)
{
	edict_t    *e;
	float      *org;

	e = G_EDICT (pr, OFS_PARM0);
	org = G_VECTOR (pr, OFS_PARM1);
	VectorCopy (org, SVFIELD (e, origin, vector));
	SV_LinkEdict (e, false);
}


void
SetMinMaxSize (progs_t * pr, edict_t *e, float *min, float *max,
			   qboolean rotate)
{
	float      *angles;
	vec3_t      rmin, rmax;
	float       bounds[2][3];
	float       xvector[2], yvector[2];
	float       a;
	vec3_t      base, transformed;
	int         i, j, k, l;

	for (i = 0; i < 3; i++)
		if (min[i] > max[i])
			PR_RunError (pr, "backwards mins/maxs");

	rotate = false;						// FIXME: implement rotation properly 
										// 
	// 
	// again

	if (!rotate) {
		VectorCopy (min, rmin);
		VectorCopy (max, rmax);
	} else {
		// find min / max for rotations
		angles = SVFIELD (e, angles, vector);

		a = angles[1] / 180 * M_PI;

		xvector[0] = cos (a);
		xvector[1] = sin (a);
		yvector[0] = -sin (a);
		yvector[1] = cos (a);

		VectorCopy (min, bounds[0]);
		VectorCopy (max, bounds[1]);

		rmin[0] = rmin[1] = rmin[2] = 9999;
		rmax[0] = rmax[1] = rmax[2] = -9999;

		for (i = 0; i <= 1; i++) {
			base[0] = bounds[i][0];
			for (j = 0; j <= 1; j++) {
				base[1] = bounds[j][1];
				for (k = 0; k <= 1; k++) {
					base[2] = bounds[k][2];

					// transform the point
					transformed[0] =
						xvector[0] * base[0] + yvector[0] * base[1];
					transformed[1] =
						xvector[1] * base[0] + yvector[1] * base[1];
					transformed[2] = base[2];

					for (l = 0; l < 3; l++) {
						if (transformed[l] < rmin[l])
							rmin[l] = transformed[l];
						if (transformed[l] > rmax[l])
							rmax[l] = transformed[l];
					}
				}
			}
		}
	}

// set derived values
	VectorCopy (rmin, SVFIELD (e, mins, vector));
	VectorCopy (rmax, SVFIELD (e, maxs, vector));
	VectorSubtract (max, min, SVFIELD (e, size, vector));

	SV_LinkEdict (e, false);
}

/*
=================
PF_setsize

the size box is rotated by the current angle

setsize (entity, minvector, maxvector)
=================
*/
void
PF_setsize (progs_t * pr)
{
	edict_t    *e;
	float      *min, *max;

	e = G_EDICT (pr, OFS_PARM0);
	min = G_VECTOR (pr, OFS_PARM1);
	max = G_VECTOR (pr, OFS_PARM2);
	SetMinMaxSize (pr, e, min, max, false);
}


/*
=================
PF_setmodel

setmodel(entity, model)
=================
*/
void
PF_setmodel (progs_t * pr)
{
	edict_t    *e;
	char       *m, **check;
	model_t    *mod;
	int         i;

	e = G_EDICT (pr, OFS_PARM0);
	m = G_STRING (pr, OFS_PARM1);

// check to see if model was properly precached
	for (i = 0, check = sv.model_precache; *check; i++, check++)
		if (!strcmp (*check, m))
			break;

	if (!*check)
		PR_RunError (pr, "no precache: %s\n", m);


	SVFIELD (e, model, string) = PR_SetString (pr, m);
	SVFIELD (e, modelindex, float) = i;				// SV_ModelIndex (m);

	mod = sv.models[(int) SVFIELD (e, modelindex, float)];	// Mod_ForName (m, true);

	if (mod)
		SetMinMaxSize (pr, e, mod->mins, mod->maxs, true);
	else
		SetMinMaxSize (pr, e, vec3_origin, vec3_origin, true);
}

/*
=================
PF_bprint

broadcast print to everyone on server

bprint(value)
=================
*/
void
PF_bprint (progs_t * pr)
{
	char       *s;

	s = PF_VarString (pr, 0);
	SV_BroadcastPrintf ("%s", s);
}

/*
=================
PF_sprint

single print to a specific client

sprint(clientent, value)
=================
*/
void
PF_sprint (progs_t * pr)
{
	char       *s;
	client_t   *client;
	int         entnum;

	entnum = G_EDICTNUM (pr, OFS_PARM0);
	s = PF_VarString (pr, 1);

	if (entnum < 1 || entnum > svs.maxclients) {
		Con_Printf ("tried to sprint to a non-client\n");
		return;
	}

	client = &svs.clients[entnum - 1];

	MSG_WriteChar (&client->message, svc_print);
	MSG_WriteString (&client->message, s);
}


/*
=================
PF_centerprint

single print to a specific client

centerprint(clientent, value)
=================
*/
void
PF_centerprint (progs_t * pr)
{
	char       *s;
	client_t   *client;
	int         entnum;

	entnum = G_EDICTNUM (pr, OFS_PARM0);
	s = PF_VarString (pr, 1);

	if (entnum < 1 || entnum > svs.maxclients) {
		Con_Printf ("tried to sprint to a non-client\n");
		return;
	}

	client = &svs.clients[entnum - 1];

	MSG_WriteChar (&client->message, svc_centerprint);
	MSG_WriteString (&client->message, s);
}


/*
=================
PF_normalize

vector normalize(vector)
=================
*/
void
PF_normalize (progs_t * pr)
{
	float      *value1;
	vec3_t      newvalue;
	float       new;

	value1 = G_VECTOR (pr, OFS_PARM0);

	new = value1[0] * value1[0] + value1[1] * value1[1] + value1[2] * value1[2];
	new = sqrt (new);

	if (new == 0)
		newvalue[0] = newvalue[1] = newvalue[2] = 0;
	else {
		new = 1 / new;
		newvalue[0] = value1[0] * new;
		newvalue[1] = value1[1] * new;
		newvalue[2] = value1[2] * new;
	}

	VectorCopy (newvalue, G_VECTOR (pr, OFS_RETURN));
}

/*
=================
PF_vlen

scalar vlen(vector)
=================
*/
void
PF_vlen (progs_t * pr)
{
	float      *value1;
	float       new;

	value1 = G_VECTOR (pr, OFS_PARM0);

	new = value1[0] * value1[0] + value1[1] * value1[1] + value1[2] * value1[2];
	new = sqrt (new);

	G_FLOAT (pr, OFS_RETURN) = new;
}

/*
=================
PF_vectoyaw

float vectoyaw(vector)
=================
*/
void
PF_vectoyaw (progs_t * pr)
{
	float      *value1;
	float       yaw;

	value1 = G_VECTOR (pr, OFS_PARM0);

	if (value1[1] == 0 && value1[0] == 0)
		yaw = 0;
	else {
		yaw = (int) (atan2 (value1[1], value1[0]) * 180 / M_PI);
		if (yaw < 0)
			yaw += 360;
	}

	G_FLOAT (pr, OFS_RETURN) = yaw;
}


/*
=================
PF_vectoangles

vector vectoangles(vector)
=================
*/
void
PF_vectoangles (progs_t * pr)
{
	float      *value1;
	float       forward;
	float       yaw, pitch;

	value1 = G_VECTOR (pr, OFS_PARM0);

	if (value1[1] == 0 && value1[0] == 0) {
		yaw = 0;
		if (value1[2] > 0)
			pitch = 90;
		else
			pitch = 270;
	} else {
		yaw = (int) (atan2 (value1[1], value1[0]) * 180 / M_PI);
		if (yaw < 0)
			yaw += 360;

		forward = sqrt (value1[0] * value1[0] + value1[1] * value1[1]);
		pitch = (int) (atan2 (value1[2], forward) * 180 / M_PI);
		if (pitch < 0)
			pitch += 360;
	}

	G_FLOAT (pr, OFS_RETURN + 0) = pitch;
	G_FLOAT (pr, OFS_RETURN + 1) = yaw;
	G_FLOAT (pr, OFS_RETURN + 2) = 0;
}

/*
=================
PF_Random

Returns a number from 0<= num < 1

random()
=================
*/
void
PF_random (progs_t * pr)
{
	float       num;

	num = (rand () & 0x7fff) / ((float) 0x7fff);

	G_FLOAT (pr, OFS_RETURN) = num;
}

/*
=================
PF_particle

particle(origin, color, count)
=================
*/
void
PF_particle (progs_t * pr)
{
	float      *org, *dir;
	float       color;
	float       count;

	org = G_VECTOR (pr, OFS_PARM0);
	dir = G_VECTOR (pr, OFS_PARM1);
	color = G_FLOAT (pr, OFS_PARM2);
	count = G_FLOAT (pr, OFS_PARM3);
	SV_StartParticle (org, dir, color, count);
}


/*
=================
PF_ambientsound

=================
*/
void
PF_ambientsound (progs_t * pr)
{
	char      **check;
	char       *samp;
	float      *pos;
	float       vol, attenuation;
	int         i, soundnum;

	pos = G_VECTOR (pr, OFS_PARM0);
	samp = G_STRING (pr, OFS_PARM1);
	vol = G_FLOAT (pr, OFS_PARM2);
	attenuation = G_FLOAT (pr, OFS_PARM3);

// check to see if samp was properly precached
	for (soundnum = 0, check = sv.sound_precache; *check; check++, soundnum++)
		if (!strcmp (*check, samp))
			break;

	if (!*check) {
		Con_Printf ("no precache: %s\n", samp);
		return;
	}
// add an svc_spawnambient command to the level signon packet

	MSG_WriteByte (&sv.signon, svc_spawnstaticsound);
	for (i = 0; i < 3; i++)
		MSG_WriteCoord (&sv.signon, pos[i]);

	MSG_WriteByte (&sv.signon, soundnum);

	MSG_WriteByte (&sv.signon, vol * 255);
	MSG_WriteByte (&sv.signon, attenuation * 64);

}

/*
=================
PF_sound

Each entity can have eight independant sound sources, like voice,
weapon, feet, etc.

Channel 0 is an auto-allocate channel, the others override anything
already running on that entity/channel pair.

An attenuation of 0 will play full volume everywhere in the level.
Larger attenuations will drop off.

=================
*/
void
PF_sound (progs_t * pr)
{
	char       *sample;
	int         channel;
	edict_t    *entity;
	int         volume;
	float       attenuation;

	entity = G_EDICT (pr, OFS_PARM0);
	channel = G_FLOAT (pr, OFS_PARM1);
	sample = G_STRING (pr, OFS_PARM2);
	volume = G_FLOAT (pr, OFS_PARM3) * 255;
	attenuation = G_FLOAT (pr, OFS_PARM4);

	if (volume < 0 || volume > 255)
		Sys_Error ("SV_StartSound: volume = %i", volume);

	if (attenuation < 0 || attenuation > 4)
		Sys_Error ("SV_StartSound: attenuation = %f", attenuation);

	if (channel < 0 || channel > 7)
		Sys_Error ("SV_StartSound: channel = %i", channel);

	SV_StartSound (entity, channel, sample, volume, attenuation);
}

/*
=================
PF_break

break()
=================
*/
void
PF_break (progs_t * pr)
{
	Con_Printf ("break statement\n");
	*(int *) -4 = 0;					// dump to debugger
//  PR_RunError ("break statement");
}

/*
=================
PF_traceline

Used for use tracing and shot targeting
Traces are blocked by bbox and exact bsp entityes, and also slide box entities
if the tryents flag is set.

traceline (vector1, vector2, tryents)
=================
*/
void
PF_traceline (progs_t * pr)
{
	float      *v1, *v2;
	trace_t     trace;
	int         nomonsters;
	edict_t    *ent;

	v1 = G_VECTOR (pr, OFS_PARM0);
	v2 = G_VECTOR (pr, OFS_PARM1);
	nomonsters = G_FLOAT (pr, OFS_PARM2);
	ent = G_EDICT (pr, OFS_PARM3);

	trace = SV_Move (v1, vec3_origin, vec3_origin, v2, nomonsters, ent);

	*sv_globals.trace_allsolid = trace.allsolid;
	*sv_globals.trace_startsolid = trace.startsolid;
	*sv_globals.trace_fraction = trace.fraction;
	*sv_globals.trace_inwater = trace.inwater;
	*sv_globals.trace_inopen = trace.inopen;
	VectorCopy (trace.endpos, *sv_globals.trace_endpos);
	VectorCopy (trace.plane.normal, *sv_globals.trace_plane_normal);
	*sv_globals.trace_plane_dist = trace.plane.dist;
	if (trace.ent)
		*sv_globals.trace_ent = EDICT_TO_PROG (pr, trace.ent);
	else
		*sv_globals.trace_ent = EDICT_TO_PROG (pr, sv.edicts);
}


#ifdef QUAKE2
extern trace_t SV_Trace_Toss (edict_t *ent, edict_t *ignore);

void
PF_TraceToss (progs_t * pr)
{
	trace_t     trace;
	edict_t    *ent;
	edict_t    *ignore;

	ent = G_EDICT (pr, OFS_PARM0);
	ignore = G_EDICT (pr, OFS_PARM1);

	trace = SV_Trace_Toss (ent, ignore);

	*sv_globals.trace_allsolid = trace.allsolid;
	*sv_globals.trace_startsolid = trace.startsolid;
	*sv_globals.trace_fraction = trace.fraction;
	*sv_globals.trace_inwater = trace.inwater;
	*sv_globals.trace_inopen = trace.inopen;
	VectorCopy (trace.endpos, *sv_globals.trace_endpos);
	VectorCopy (trace.plane.normal, *sv_globals.trace_plane_normal);
	*sv_globals.trace_plane_dist = trace.plane.dist;
	if (trace.ent)
		*sv_globals.trace_ent = EDICT_TO_PROG (pr, trace.ent);
	else
		*sv_globals.trace_ent = EDICT_TO_PROG (pr, sv.edicts);
}
#endif


/*
=================
PF_checkpos

Returns true if the given entity can move to the given position from it's
current position by walking or rolling.
FIXME: make work...
scalar checkpos (entity, vector)
=================
*/
void
PF_checkpos (progs_t * pr)
{
}

//============================================================================

byte        checkpvs[MAX_MAP_LEAFS / 8];

int
PF_newcheckclient (progs_t * pr, int check)
{
	int         i;
	byte       *pvs;
	edict_t    *ent;
	mleaf_t    *leaf;
	vec3_t      org;

// cycle to the next one

	if (check < 1)
		check = 1;
	if (check > svs.maxclients)
		check = svs.maxclients;

	if (check == svs.maxclients)
		i = 1;
	else
		i = check + 1;

	for (;; i++) {
		if (i == svs.maxclients + 1)
			i = 1;

		ent = EDICT_NUM (pr, i);

		if (i == check)
			break;						// didn't find anything else

		if (ent->free)
			continue;
		if (SVFIELD (ent, health, float) <= 0)
			continue;
		if ((int) SVFIELD (ent, flags, float) & FL_NOTARGET)
			continue;

		// anything that is a client, or has a client as an enemy
		break;
	}

// get the PVS for the entity
	VectorAdd (SVFIELD (ent, origin, vector), SVFIELD (ent, view_ofs, vector), org);
	leaf = Mod_PointInLeaf (org, sv.worldmodel);
	pvs = Mod_LeafPVS (leaf, sv.worldmodel);
	memcpy (checkpvs, pvs, (sv.worldmodel->numleafs + 7) >> 3);

	return i;
}

/*
=================
PF_checkclient

Returns a client (or object that has a client enemy) that would be a
valid target.

If there are more than one valid options, they are cycled each frame

If (self.origin + self.viewofs) is not in the PVS of the current target,
it is not returned at all.

name checkclient ()
=================
*/
#define	MAX_CHECK	16
int         c_invis, c_notvis;
void
PF_checkclient (progs_t * pr)
{
	edict_t    *ent, *self;
	mleaf_t    *leaf;
	int         l;
	vec3_t      view;

// find a new check if on a new frame
	if (sv.time - sv.lastchecktime >= 0.1) {
		sv.lastcheck = PF_newcheckclient (pr, sv.lastcheck);
		sv.lastchecktime = sv.time;
	}
// return check if it might be visible  
	ent = EDICT_NUM (pr, sv.lastcheck);
	if (ent->free || SVFIELD (ent, health, float) <= 0) {
		RETURN_EDICT (pr, sv.edicts);
		return;
	}
// if current entity can't possibly see the check entity, return 0
	self = PROG_TO_EDICT (pr, *sv_globals.self);
	VectorAdd (SVFIELD (self, origin, vector), SVFIELD (self, view_ofs, vector), view);
	leaf = Mod_PointInLeaf (view, sv.worldmodel);
	l = (leaf - sv.worldmodel->leafs) - 1;
	if ((l < 0) || !(checkpvs[l >> 3] & (1 << (l & 7)))) {
		c_notvis++;
		RETURN_EDICT (pr, sv.edicts);
		return;
	}
// might be able to see it
	c_invis++;
	RETURN_EDICT (pr, ent);
}

//============================================================================


/*
=================
PF_stuffcmd

Sends text over to the client's execution buffer

stuffcmd (clientent, value)
=================
*/
void
PF_stuffcmd (progs_t * pr)
{
	int         entnum;
	char       *str;
	client_t   *old;

	entnum = G_EDICTNUM (pr, OFS_PARM0);
	if (entnum < 1 || entnum > svs.maxclients)
		PR_RunError (pr, "Parm 0 not a client");
	str = G_STRING (pr, OFS_PARM1);

	old = host_client;
	host_client = &svs.clients[entnum - 1];
	Host_ClientCommands ("%s", str);
	host_client = old;
}

/*
=================
PF_localcmd

Sends text over to the client's execution buffer

localcmd (string)
=================
*/
void
PF_localcmd (progs_t * pr)
{
	char       *str;

	str = G_STRING (pr, OFS_PARM0);
	Cbuf_AddText (str);
}

/*
=================
PF_cvar

float cvar (string)
=================
*/
void
PF_cvar (progs_t * pr)
{
	char       *str;

	str = G_STRING (pr, OFS_PARM0);

	G_FLOAT (pr, OFS_RETURN) = Cvar_VariableValue (str);
}

/*
=================
PF_cvar_set

float cvar (string)
=================
*/
void
PF_cvar_set (progs_t * pr)
{
	char       *var_name, *val;
	cvar_t     *var;

	var_name = G_STRING (pr, OFS_PARM0);
	val = G_STRING (pr, OFS_PARM1);
	var = Cvar_FindVar (var_name);
	if (!var)
		var = Cvar_FindAlias (var_name);
	if (!var) {
		// FIXME: make Con_DPrint?
		Con_Printf ("PF_cvar_set: variable %s not found\n", var_name);
		return;
	}

	Cvar_Set (var, val);
}

/*
=================
PF_findradius

Returns a chain of entities that have origins within a spherical area

findradius (origin, radius)
=================
*/
void
PF_findradius (progs_t * pr)
{
	edict_t    *ent, *chain;
	float       rad;
	float      *org;
	vec3_t      eorg;
	int         i, j;

	chain = (edict_t *) sv.edicts;

	org = G_VECTOR (pr, OFS_PARM0);
	rad = G_FLOAT (pr, OFS_PARM1);

	ent = NEXT_EDICT (pr, sv.edicts);
	for (i = 1; i < sv.num_edicts; i++, ent = NEXT_EDICT (pr, ent)) {
		if (ent->free)
			continue;
		if (SVFIELD (ent, solid, float) == SOLID_NOT)
			continue;
		for (j = 0; j < 3; j++)
			eorg[j] =
				org[j] - (SVFIELD (ent, origin, vector)[j] +
						  (SVFIELD (ent, mins, vector)[j] + SVFIELD (ent, maxs, vector)[j]) * 0.5);
		if (Length (eorg) > rad)
			continue;

		SVFIELD (ent, chain, entity) = EDICT_TO_PROG (pr, chain);
		chain = ent;
	}

	RETURN_EDICT (pr, chain);
}


/*
=========
PF_dprint
=========
*/
void
PF_dprint (progs_t * pr)
{
	Con_DPrintf ("%s", PF_VarString (pr, 0));
}

char        pr_string_temp[128];

void
PF_ftos (progs_t * pr)
{
	float       v;

	v = G_FLOAT (pr, OFS_PARM0);

	if (v == (int) v)
		snprintf (pr_string_temp, sizeof (pr_string_temp), "%d", (int) v);
	else
		snprintf (pr_string_temp, sizeof (pr_string_temp), "%5.1f", v);
	G_INT (pr, OFS_RETURN) = PR_SetString (pr, pr_string_temp);
}

void
PF_fabs (progs_t * pr)
{
	float       v;

	v = G_FLOAT (pr, OFS_PARM0);
	G_FLOAT (pr, OFS_RETURN) = fabs (v);
}

void
PF_vtos (progs_t * pr)
{
	snprintf (pr_string_temp, sizeof (pr_string_temp),
			  "'%5.1f %5.1f %5.1f'", G_VECTOR (pr, OFS_PARM0)[0], G_VECTOR (pr,
																			OFS_PARM0)
			  [1], G_VECTOR (pr, OFS_PARM0)[2]);
	G_INT (pr, OFS_RETURN) = PR_SetString (pr, pr_string_temp);
}

#ifdef QUAKE2
void
PF_etos (progs_t * pr)
{
	snprintf (pr_string_temp, sizeof (pr_string_temp), "entity %i",
			  G_EDICTNUM (pr, OFS_PARM0));
	G_INT (pr, OFS_RETURN) = PR_SetString (pr, pr_string_temp);
}
#endif

void
PF_Spawn (progs_t * pr)
{
	edict_t    *ed;

	ed = ED_Alloc (pr);
	RETURN_EDICT (pr, ed);
}

void
PF_Remove (progs_t * pr)
{
	edict_t    *ed;

	ed = G_EDICT (pr, OFS_PARM0);
	ED_Free (pr, ed);
}


// entity (entity start, .string field, string match) find = #5;
void
PF_Find (progs_t * pr)
#ifdef QUAKE2
{
	int         e;
	int         f;
	char       *s, *t;
	edict_t    *ed;
	edict_t    *first;
	edict_t    *second;
	edict_t    *last;

	first = second = last = (edict_t *) sv.edicts;
	e = G_EDICTNUM (pr, OFS_PARM0);
	f = G_INT (pr, OFS_PARM1);
	s = G_STRING (pr, OFS_PARM2);
	if (!s)
		PR_RunError (pr, "PF_Find: bad search string");

	for (e++; e < sv.num_edicts; e++) {
		ed = EDICT_NUM (pr, e);
		if (ed->free)
			continue;
		t = E_STRING (pr, ed, f);
		if (!t)
			continue;
		if (!strcmp (t, s)) {
			if (first == (edict_t *) sv.edicts)
				first = ed;
			else if (second == (edict_t *) sv.edicts)
				second = ed;
			SVFIELD (ed, chain, entity) = EDICT_TO_PROG (pr, last);
			last = ed;
		}
	}

	if (first != last) {
		if (last != second)
			SVFIELD (first, chain, entity) = SVFIELD (last, chain, entity);
		else
			SVFIELD (first, chain, entity) = EDICT_TO_PROG (pr, last);
		SVFIELD (last, chain, entity) = EDICT_TO_PROG (pr, (edict_t *) sv.edicts);
		if (second && second != last)
			SVFIELD (second, chain, entity) = EDICT_TO_PROG (pr, last);
	}
	RETURN_EDICT (pr, first);
}
#else
{
	int         e;
	int         f;
	char       *s, *t;
	edict_t    *ed;

	e = G_EDICTNUM (pr, OFS_PARM0);
	f = G_INT (pr, OFS_PARM1);
	s = G_STRING (pr, OFS_PARM2);
	if (!s)
		PR_RunError (pr, "PF_Find: bad search string");

	for (e++; e < sv.num_edicts; e++) {
		ed = EDICT_NUM (pr, e);
		if (ed->free)
			continue;
		t = E_STRING (pr, ed, f);
		if (!t)
			continue;
		if (!strcmp (t, s)) {
			RETURN_EDICT (pr, ed);
			return;
		}
	}

	RETURN_EDICT (pr, sv.edicts);
}
#endif

void
PR_CheckEmptyString (progs_t * pr, char *s)
{
	if (s[0] <= ' ')
		PR_RunError (pr, "Bad string");
}

void
PF_precache_file (progs_t * pr)
{										// precache_file is only used to copy 
										// 
	// 
	// files with qcc, it does nothing
	G_INT (pr, OFS_RETURN) = G_INT (pr, OFS_PARM0);
}

void
PF_precache_sound (progs_t * pr)
{
	char       *s;
	int         i;

	if (sv.state != ss_loading)
		PR_RunError
			(pr, "PF_Precache_*: Precache can only be done in spawn functions");

	s = G_STRING (pr, OFS_PARM0);
	G_INT (pr, OFS_RETURN) = G_INT (pr, OFS_PARM0);
	PR_CheckEmptyString (pr, s);

	for (i = 0; i < MAX_SOUNDS; i++) {
		if (!sv.sound_precache[i]) {
			sv.sound_precache[i] = s;
			return;
		}
		if (!strcmp (sv.sound_precache[i], s))
			return;
	}
	PR_RunError (pr, "PF_precache_sound: overflow");
}

void
PF_precache_model (progs_t * pr)
{
	char       *s;
	int         i;

	if (sv.state != ss_loading)
		PR_RunError
			(pr, "PF_Precache_*: Precache can only be done in spawn functions");

	s = G_STRING (pr, OFS_PARM0);
	G_INT (pr, OFS_RETURN) = G_INT (pr, OFS_PARM0);
	PR_CheckEmptyString (pr, s);

	for (i = 0; i < MAX_MODELS; i++) {
		if (!sv.model_precache[i]) {
			sv.model_precache[i] = s;
			sv.models[i] = Mod_ForName (s, true);
			return;
		}
		if (!strcmp (sv.model_precache[i], s))
			return;
	}
	PR_RunError (pr, "PF_precache_model: overflow");
}


void
PF_coredump (progs_t * pr)
{
	ED_PrintEdicts (pr);
}

void
PF_traceon (progs_t * pr)
{
	pr->pr_trace = true;
}

void
PF_traceoff (progs_t * pr)
{
	pr->pr_trace = false;
}

void
PF_eprint (progs_t * pr)
{
	ED_PrintNum (pr, G_EDICTNUM (pr, OFS_PARM0));
}

/*
===============
PF_walkmove

float(float yaw, float dist) walkmove
===============
*/
void
PF_walkmove (progs_t * pr)
{
	edict_t    *ent;
	float       yaw, dist;
	vec3_t      move;
	dfunction_t *oldf;
	int         oldself;

	ent = PROG_TO_EDICT (pr, *sv_globals.self);
	yaw = G_FLOAT (pr, OFS_PARM0);
	dist = G_FLOAT (pr, OFS_PARM1);

	if (!((int) SVFIELD (ent, flags, float) & (FL_ONGROUND | FL_FLY | FL_SWIM))) {
		G_FLOAT (pr, OFS_RETURN) = 0;
		return;
	}

	yaw = yaw * M_PI * 2 / 360;

	move[0] = cos (yaw) * dist;
	move[1] = sin (yaw) * dist;
	move[2] = 0;

// save program state, because SV_movestep may call other progs
	oldf = pr->pr_xfunction;
	oldself = *sv_globals.self;

	G_FLOAT (pr, OFS_RETURN) = SV_movestep (ent, move, true);


// restore program state
	pr->pr_xfunction = oldf;
	*sv_globals.self = oldself;
}

/*
===============
PF_droptofloor

void() droptofloor
===============
*/
void
PF_droptofloor (progs_t * pr)
{
	edict_t    *ent;
	vec3_t      end;
	trace_t     trace;

	ent = PROG_TO_EDICT (pr, *sv_globals.self);

	VectorCopy (SVFIELD (ent, origin, vector), end);
	end[2] -= 256;

	trace =
		SV_Move (SVFIELD (ent, origin, vector), SVFIELD (ent, mins, vector), SVFIELD (ent, maxs, vector), end, false,
				 ent);

	if (trace.fraction == 1 || trace.allsolid)
		G_FLOAT (pr, OFS_RETURN) = 0;
	else {
		VectorCopy (trace.endpos, SVFIELD (ent, origin, vector));
		SV_LinkEdict (ent, false);
		SVFIELD (ent, flags, float) = (int) SVFIELD (ent, flags, float) | FL_ONGROUND;
		SVFIELD (ent, groundentity, entity) = EDICT_TO_PROG (pr, trace.ent);
		G_FLOAT (pr, OFS_RETURN) = 1;
	}
}

/*
===============
PF_lightstyle

void(float style, string value) lightstyle
===============
*/
void
PF_lightstyle (progs_t * pr)
{
	int         style;
	char       *val;
	client_t   *client;
	int         j;

	style = G_FLOAT (pr, OFS_PARM0);
	val = G_STRING (pr, OFS_PARM1);

// change the string in sv
	sv.lightstyles[style] = val;

// send message to all clients on this server
	if (sv.state != ss_active)
		return;

	for (j = 0, client = svs.clients; j < svs.maxclients; j++, client++)
		if (client->active || client->spawned) {
			MSG_WriteChar (&client->message, svc_lightstyle);
			MSG_WriteChar (&client->message, style);
			MSG_WriteString (&client->message, val);
		}
}

void
PF_rint (progs_t * pr)
{
	float       f;

	f = G_FLOAT (pr, OFS_PARM0);
	if (f > 0)
		G_FLOAT (pr, OFS_RETURN) = (int) (f + 0.5);
	else
		G_FLOAT (pr, OFS_RETURN) = (int) (f - 0.5);
}

void
PF_floor (progs_t * pr)
{
	G_FLOAT (pr, OFS_RETURN) = floor (G_FLOAT (pr, OFS_PARM0));
}

void
PF_ceil (progs_t * pr)
{
	G_FLOAT (pr, OFS_RETURN) = ceil (G_FLOAT (pr, OFS_PARM0));
}


/*
=============
PF_checkbottom
=============
*/
void
PF_checkbottom (progs_t * pr)
{
	edict_t    *ent;

	ent = G_EDICT (pr, OFS_PARM0);

	G_FLOAT (pr, OFS_RETURN) = SV_CheckBottom (ent);
}

/*
=============
PF_pointcontents
=============
*/
void
PF_pointcontents (progs_t * pr)
{
	float      *v;

	v = G_VECTOR (pr, OFS_PARM0);

	G_FLOAT (pr, OFS_RETURN) = SV_PointContents (v);
}

/*
=============
PF_nextent

entity nextent(entity)
=============
*/
void
PF_nextent (progs_t * pr)
{
	int         i;
	edict_t    *ent;

	i = G_EDICTNUM (pr, OFS_PARM0);
	while (1) {
		i++;
		if (i == sv.num_edicts) {
			RETURN_EDICT (pr, sv.edicts);
			return;
		}
		ent = EDICT_NUM (pr, i);
		if (!ent->free) {
			RETURN_EDICT (pr, ent);
			return;
		}
	}
}

/*
=============
PF_aim

Pick a vector for the player to shoot along
vector aim(entity, missilespeed)
=============
*/
cvar_t     *sv_aim;
void
PF_aim (progs_t * pr)
{
	edict_t    *ent, *check, *bestent;
	vec3_t      start, dir, end, bestdir;
	int         i, j;
	trace_t     tr;
	float       dist, bestdist;
	float       speed;

	ent = G_EDICT (pr, OFS_PARM0);
	speed = G_FLOAT (pr, OFS_PARM1);

	VectorCopy (SVFIELD (ent, origin, vector), start);
	start[2] += 20;

// try sending a trace straight
	VectorCopy (*sv_globals.v_forward, dir);
	VectorMA (start, 2048, dir, end);
	tr = SV_Move (start, vec3_origin, vec3_origin, end, false, ent);
	if (tr.ent && SVFIELD (tr.ent, takedamage, float) == DAMAGE_AIM
		&& (!teamplay->int_val || SVFIELD (ent, team, float) <= 0
			|| SVFIELD (ent, team, float) != SVFIELD (tr.ent, team, float))) {
		VectorCopy (*sv_globals.v_forward, G_VECTOR (pr, OFS_RETURN));
		return;
	}
// try all possible entities
	VectorCopy (dir, bestdir);
	bestdist = sv_aim->value;
	bestent = NULL;

	check = NEXT_EDICT (pr, sv.edicts);
	for (i = 1; i < sv.num_edicts; i++, check = NEXT_EDICT (pr, check)) {
		if (SVFIELD (check, takedamage, float) != DAMAGE_AIM)
			continue;
		if (check == ent)
			continue;
		if (teamplay->int_val && SVFIELD (ent, team, float) > 0
			&& SVFIELD (ent, team, float) == SVFIELD (check, team, float)) continue;	// don't aim at
		// teammate
		for (j = 0; j < 3; j++)
			end[j] = SVFIELD (check, origin, vector)[j]
				+ 0.5 * (SVFIELD (check, mins, vector)[j] + SVFIELD (check, maxs, vector)[j]);
		VectorSubtract (end, start, dir);
		VectorNormalize (dir);
		dist = DotProduct (dir, *sv_globals.v_forward);
		if (dist < bestdist)
			continue;					// to far to turn
		tr = SV_Move (start, vec3_origin, vec3_origin, end, false, ent);
		if (tr.ent == check) {			// can shoot at this one
			bestdist = dist;
			bestent = check;
		}
	}

	if (bestent) {
		VectorSubtract (SVFIELD (bestent, origin, vector), SVFIELD (ent, origin, vector), dir);
		dist = DotProduct (dir, *sv_globals.v_forward);
		VectorScale (*sv_globals.v_forward, dist, end);
		end[2] = dir[2];
		VectorNormalize (end);
		VectorCopy (end, G_VECTOR (pr, OFS_RETURN));
	} else {
		VectorCopy (bestdir, G_VECTOR (pr, OFS_RETURN));
	}
}

/*
==============
PF_changeyaw

This was a major timewaster in progs, so it was converted to C
==============
*/
void
PF_changeyaw (progs_t * pr)
{
	edict_t    *ent;
	float       ideal, current, move, speed;

	ent = PROG_TO_EDICT (pr, *sv_globals.self);
	current = anglemod (SVFIELD (ent, angles, vector)[1]);
	ideal = SVFIELD (ent, ideal_yaw, float);
	speed = SVFIELD (ent, yaw_speed, float);

	if (current == ideal)
		return;
	move = ideal - current;
	if (ideal > current) {
		if (move >= 180)
			move = move - 360;
	} else {
		if (move <= -180)
			move = move + 360;
	}
	if (move > 0) {
		if (move > speed)
			move = speed;
	} else {
		if (move < -speed)
			move = -speed;
	}

	SVFIELD (ent, angles, vector)[1] = anglemod (current + move);
}

#ifdef QUAKE2
/*
==============
PF_changepitch
==============
*/
void
PF_changepitch (progs_t * pr)
{
	edict_t    *ent;
	float       ideal, current, move, speed;

	ent = G_EDICT (pr, OFS_PARM0);
	current = anglemod (SVFIELD (ent, angles, vector)[0]);
	ideal = SVFIELD (ent, idealpitch, float);
	speed = SVFIELD (ent, pitch_speed, float);

	if (current == ideal)
		return;
	move = ideal - current;
	if (ideal > current) {
		if (move >= 180)
			move = move - 360;
	} else {
		if (move <= -180)
			move = move + 360;
	}
	if (move > 0) {
		if (move > speed)
			move = speed;
	} else {
		if (move < -speed)
			move = -speed;
	}

	SVFIELD (ent, angles, vector)[0] = anglemod (current + move);
}
#endif

/*
===============================================================================

MESSAGE WRITING

===============================================================================
*/

#define	MSG_BROADCAST	0				// unreliable to all
#define	MSG_ONE			1				// reliable to one (msg_entity)
#define	MSG_ALL			2				// reliable to all
#define	MSG_INIT		3				// write to the init string

sizebuf_t  *
WriteDest (progs_t * pr)
{
	int         entnum;
	int         dest;
	edict_t    *ent;

	dest = G_FLOAT (pr, OFS_PARM0);
	switch (dest) {
		case MSG_BROADCAST:
		return &sv.datagram;

		case MSG_ONE:
		ent = PROG_TO_EDICT (pr, *sv_globals.msg_entity);
		entnum = NUM_FOR_EDICT (pr, ent);
		if (entnum < 1 || entnum > svs.maxclients)
			PR_RunError (pr, "WriteDest: not a client");
		return &svs.clients[entnum - 1].message;

		case MSG_ALL:
		return &sv.reliable_datagram;

		case MSG_INIT:
		return &sv.signon;

		default:
		PR_RunError (pr, "WriteDest: bad destination");
		break;
	}

	return NULL;
}

void
PF_WriteByte (progs_t * pr)
{
	MSG_WriteByte (WriteDest (pr), G_FLOAT (pr, OFS_PARM1));
}

void
PF_WriteChar (progs_t * pr)
{
	MSG_WriteChar (WriteDest (pr), G_FLOAT (pr, OFS_PARM1));
}

void
PF_WriteShort (progs_t * pr)
{
	MSG_WriteShort (WriteDest (pr), G_FLOAT (pr, OFS_PARM1));
}

void
PF_WriteLong (progs_t * pr)
{
	MSG_WriteLong (WriteDest (pr), G_FLOAT (pr, OFS_PARM1));
}

void
PF_WriteAngle (progs_t * pr)
{
	MSG_WriteAngle (WriteDest (pr), G_FLOAT (pr, OFS_PARM1));
}

void
PF_WriteCoord (progs_t * pr)
{
	MSG_WriteCoord (WriteDest (pr), G_FLOAT (pr, OFS_PARM1));
}

void
PF_WriteString (progs_t * pr)
{
	MSG_WriteString (WriteDest (pr), G_STRING (pr, OFS_PARM1));
}


void
PF_WriteEntity (progs_t * pr)
{
	MSG_WriteShort (WriteDest (pr), G_EDICTNUM (pr, OFS_PARM1));
}

//=============================================================================

int         SV_ModelIndex (char *name);

void
PF_makestatic (progs_t * pr)
{
	edict_t    *ent;
	int         i;

	ent = G_EDICT (pr, OFS_PARM0);

	MSG_WriteByte (&sv.signon, svc_spawnstatic);

	MSG_WriteByte (&sv.signon, SV_ModelIndex (PR_GetString (pr, SVFIELD (ent, model, string))));

	MSG_WriteByte (&sv.signon, SVFIELD (ent, frame, float));
	MSG_WriteByte (&sv.signon, SVFIELD (ent, colormap, float));
	MSG_WriteByte (&sv.signon, SVFIELD (ent, skin, float));
	for (i = 0; i < 3; i++) {
		MSG_WriteCoord (&sv.signon, SVFIELD (ent, origin, vector)[i]);
		MSG_WriteAngle (&sv.signon, SVFIELD (ent, angles, vector)[i]);
	}

// throw the entity away now
	ED_Free (pr, ent);
}

//=============================================================================

/*
==============
PF_setspawnparms
==============
*/
void
PF_setspawnparms (progs_t * pr)
{
	edict_t    *ent;
	int         i;
	client_t   *client;

	ent = G_EDICT (pr, OFS_PARM0);
	i = NUM_FOR_EDICT (pr, ent);
	if (i < 1 || i > svs.maxclients)
		PR_RunError (pr, "Entity is not a client");

	// copy spawn parms out of the client_t
	client = svs.clients + (i - 1);

	for (i = 0; i < NUM_SPAWN_PARMS; i++)
		sv_globals.parms[i] = client->spawn_parms[i];
}

/*
==============
PF_changelevel
==============
*/
void
PF_changelevel (progs_t * pr)
{
#ifdef QUAKE2
	char       *s1, *s2;

	if (svs.changelevel_issued)
		return;
	svs.changelevel_issued = true;

	s1 = G_STRING (pr, OFS_PARM0);
	s2 = G_STRING (pr, OFS_PARM1);

	if ((int) *sv_globals.
		serverflags & (SFL_NEW_UNIT | SFL_NEW_EPISODE))
			Cbuf_AddText (va ("changelevel %s %s\n", s1, s2));
	else
		Cbuf_AddText (va ("changelevel2 %s %s\n", s1, s2));
#else
	char       *s;

// make sure we don't issue two changelevels
	if (svs.changelevel_issued)
		return;
	svs.changelevel_issued = true;

	s = G_STRING (pr, OFS_PARM0);
	Cbuf_AddText (va ("changelevel %s\n", s));
#endif
}


#ifdef QUAKE2

#define	CONTENT_WATER	-3
#define CONTENT_SLIME	-4
#define CONTENT_LAVA	-5

#define FL_IMMUNE_WATER	131072
#define	FL_IMMUNE_SLIME	262144
#define FL_IMMUNE_LAVA	524288

#define	CHAN_VOICE	2
#define	CHAN_BODY	4

#define	ATTN_NORM	1

void
PF_WaterMove (progs_t * pr)
{
	edict_t    *self;
	int         flags;
	int         waterlevel;
	int         watertype;
	float       drownlevel;
	float       damage = 0.0;

	self = PROG_TO_EDICT (pr, *sv_globals.self);

	if (SVFIELD (self, movetype, float) == MOVETYPE_NOCLIP) {
		SVFIELD (self, air_finished, float) = sv.time + 12;
		G_FLOAT (pr, OFS_RETURN) = damage;
		return;
	}

	if (SVFIELD (self, health, float) < 0) {
		G_FLOAT (pr, OFS_RETURN) = damage;
		return;
	}

	if (SVFIELD (self, deadflag, float) == DEAD_NO)
		drownlevel = 3;
	else
		drownlevel = 1;

	flags = (int) SVFIELD (self, flags, float);
	waterlevel = (int) SVFIELD (self, waterlevel, float);
	watertype = (int) SVFIELD (self, watertype, float);

	if (!(flags & (FL_IMMUNE_WATER + FL_GODMODE)))
		if (((flags & FL_SWIM) && (waterlevel < drownlevel))
			|| (waterlevel >= drownlevel)) {
			if (SVFIELD (self, air_finished, float) < sv.time)
				if (SVFIELD (self, pain_finished, float) < sv.time) {
					SVFIELD (self, dmg, float) = SVFIELD (self, dmg, float) + 2;
					if (SVFIELD (self, dmg, float) > 15)
						SVFIELD (self, dmg, float) = 10;
//                  T_Damage (self, world, world, self.dmg, 0, FALSE);
					damage = SVFIELD (self, dmg, float);
					SVFIELD (self, pain_finished, float) = sv.time + 1.0;
				}
		} else {
			if (SVFIELD (self, air_finished, float) < sv.time)
//              sound (self, CHAN_VOICE, "player/gasp2.wav", 1, ATTN_NORM);
				SV_StartSound (self, CHAN_VOICE, "player/gasp2.wav", 255,
							   ATTN_NORM);
			else if (SVFIELD (self, air_finished, float) < sv.time + 9)
//              sound (self, CHAN_VOICE, "player/gasp1.wav", 1, ATTN_NORM);
				SV_StartSound (self, CHAN_VOICE, "player/gasp1.wav", 255,
							   ATTN_NORM);
			SVFIELD (self, air_finished, float) = sv.time + 12.0;
			SVFIELD (self, dmg, float) = 2;
		}

	if (!waterlevel) {
		if (flags & FL_INWATER) {
			// play leave water sound
//          sound (self, CHAN_BODY, "misc/outwater.wav", 1, ATTN_NORM);
			SV_StartSound (self, CHAN_BODY, "misc/outwater.wav", 255,
						   ATTN_NORM);
			SVFIELD (self, flags, float) = (float) (flags & ~FL_INWATER);
		}
		SVFIELD (self, air_finished, float) = sv.time + 12.0;
		G_FLOAT (pr, OFS_RETURN) = damage;
		return;
	}

	if (watertype == CONTENT_LAVA) {	// do damage
		if (!(flags & (FL_IMMUNE_LAVA + FL_GODMODE)))
			if (SVFIELD (self, dmgtime, float) < sv.time) {
				if (SVFIELD (self, radsuit_finished, float) < sv.time)
					SVFIELD (self, dmgtime, float) = sv.time + 0.2;
				else
					SVFIELD (self, dmgtime, float) = sv.time + 1.0;
//              T_Damage (self, world, world, 10*self.waterlevel, 0, TRUE);
				damage = (float) (10 * waterlevel);
			}
	} else if (watertype == CONTENT_SLIME) {	// do damage
		if (!(flags & (FL_IMMUNE_SLIME + FL_GODMODE)))
			if (SVFIELD (self, dmgtime, float) < sv.time
				&& SVFIELD (self, radsuit_finished, float) < sv.time) {
				SVFIELD (self, dmgtime, float) = sv.time + 1.0;
//              T_Damage (self, world, world, 4*self.waterlevel, 0, TRUE);
				damage = (float) (4 * waterlevel);
			}
	}

	if (!(flags & FL_INWATER)) {

// player enter water sound
		if (watertype == CONTENT_LAVA)
//          sound (self, CHAN_BODY, "player/inlava.wav", 1, ATTN_NORM);
			SV_StartSound (self, CHAN_BODY, "player/inlava.wav", 255,
						   ATTN_NORM);
		if (watertype == CONTENT_WATER)
//          sound (self, CHAN_BODY, "player/inh2o.wav", 1, ATTN_NORM);
			SV_StartSound (self, CHAN_BODY, "player/inh2o.wav", 255, ATTN_NORM);
		if (watertype == CONTENT_SLIME)
//          sound (self, CHAN_BODY, "player/slimbrn2.wav", 1, ATTN_NORM);
			SV_StartSound (self, CHAN_BODY, "player/slimbrn2.wav", 255,
						   ATTN_NORM);

		SVFIELD (self, flags, float) = (float) (flags | FL_INWATER);
		SVFIELD (self, dmgtime, float) = 0;
	}

	if (!(flags & FL_WATERJUMP)) {
//      self.velocity = self.velocity - 0.8*self.waterlevel*frametime*self.velocity;
		VectorMA (SVFIELD (self, velocity, vector),
				  -0.8 * SVFIELD (self, waterlevel, float) * host_frametime,
				  SVFIELD (self, velocity, vector), SVFIELD (self, velocity, vector));
	}

	G_FLOAT (pr, OFS_RETURN) = damage;
}


void
PF_sin (progs_t * pr)
{
	G_FLOAT (pr, OFS_RETURN) = sin (G_FLOAT (pr, OFS_PARM0));
}

void
PF_cos (progs_t * pr)
{
	G_FLOAT (pr, OFS_RETURN) = cos (G_FLOAT (pr, OFS_PARM0));
}

void
PF_sqrt (progs_t * pr)
{
	G_FLOAT (pr, OFS_RETURN) = sqrt (G_FLOAT (pr, OFS_PARM0));
}

#endif

void
PF_Fixme (progs_t * pr)
{
	PR_RunError (pr, "unimplemented bulitin");
}


void
PF_checkextension (progs_t *pr)
{
	G_FLOAT(pr, OFS_RETURN) = 0; //FIXME make this function actually useful :P
}


builtin_t   sv_builtins[] = {
	PF_Fixme,
	PF_makevectors,						// void(entity e)   makevectors
	// = #1;
	PF_setorigin,						// void(entity e, vector o) setorigin
	// = #2;
	PF_setmodel,						// void(entity e, string m) setmodel
	// = #3;
	PF_setsize,							// void(entity e, vector min, vector
	// max) setsize = #4;
	PF_Fixme,							// void(entity e, vector min, vector
	// max) setabssize = #5;
	PF_break,							// void() break                     = 
										// 
	// 
	// #6;
	PF_random,							// float() random
	// = #7;
	PF_sound,							// void(entity e, float chan, string
	// samp) sound = #8;
	PF_normalize,						// vector(vector v) normalize
	// = #9;
	PF_error,							// void(string e) error             = 
										// 
	// 
	// #10;
	PF_objerror,						// void(string e) objerror
	// = #11;
	PF_vlen,							// float(vector v) vlen             = 
										// 
	// 
	// #12;
	PF_vectoyaw,						// float(vector v) vectoyaw     =
	// #13;
	PF_Spawn,							// entity() spawn
	// = #14;
	PF_Remove,							// void(entity e) remove
	// = #15;
	PF_traceline,						// float(vector v1, vector v2, float
	// tryents) traceline = #16;
	PF_checkclient,						// entity() clientlist
	// = #17;
	PF_Find,							// entity(entity start, .string fld,
	// string match) find = #18;
	PF_precache_sound,					// void(string s) precache_sound
	// = #19;
	PF_precache_model,					// void(string s) precache_model
	// = #20;
	PF_stuffcmd,						// void(entity client, string
	// s)stuffcmd = #21;
	PF_findradius,						// entity(vector org, float rad)
	// findradius = #22;
	PF_bprint,							// void(string s) bprint
	// = #23;
	PF_sprint,							// void(entity client, string s)
	// sprint = #24;
	PF_dprint,							// void(string s) dprint
	// = #25;
	PF_ftos,							// void(string s) ftos              = 
										// 
	// 
	// #26;
	PF_vtos,							// void(string s) vtos              = 
										// 
	// 
	// #27;
	PF_coredump,
	PF_traceon,
	PF_traceoff,
	PF_eprint,							// void(entity e) debug print an
	// entire entity
	PF_walkmove,						// float(float yaw, float dist)
	// walkmove
	PF_Fixme,							// float(float yaw, float dist)
	// walkmove
	PF_droptofloor,
	PF_lightstyle,
	PF_rint,
	PF_floor,
	PF_ceil,
	PF_Fixme,
	PF_checkbottom,
	PF_pointcontents,
	PF_Fixme,
	PF_fabs,
	PF_aim,
	PF_cvar,
	PF_localcmd,
	PF_nextent,
	PF_particle,
	PF_changeyaw,
	PF_Fixme,
	PF_vectoangles,

	PF_WriteByte,
	PF_WriteChar,
	PF_WriteShort,
	PF_WriteLong,
	PF_WriteCoord,
	PF_WriteAngle,
	PF_WriteString,
	PF_WriteEntity,

#ifdef QUAKE2
	PF_sin,
	PF_cos,
	PF_sqrt,
	PF_changepitch,
	PF_TraceToss,
	PF_etos,
	PF_WaterMove,
#else
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
#endif

	SV_MoveToGoal,
	PF_precache_file,
	PF_makestatic,

	PF_changelevel,
	PF_Fixme,

	PF_cvar_set,
	PF_centerprint,

	PF_ambientsound,

	PF_precache_model,
	PF_precache_sound,					// precache_sound2 is different only
	// for qcc
	PF_precache_file,

	PF_setspawnparms,

	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_Fixme,
	PF_checkextension,
};

int         sv_numbuiltins = sizeof (sv_builtins) / sizeof (sv_builtins[0]);
